@@ -30,6 +30,7 @@ Driver API Errors
     + CypherTypeError
     + ConstraintError
     + AuthError
+      + TokenExpired
     + Forbidden
   + DatabaseError
   + TransientError
@@ -129,6 +130,21 @@ class Neo4jError(Exception):
     def invalidates_all_connections(self):
         return self.code == "Neo.ClientError.Security.AuthorizationExpired"
 
+    def is_fatal_during_discovery(self):
+        # checks if the code is an error that is caused by the client. In this
+        # case the driver should fail fast during discovery.
+        if not isinstance(self.code, str):
+            return False
+        if self.code in ("Neo.ClientError.Database.DatabaseNotFound",
+                         "Neo.ClientError.Transaction.InvalidBookmark",
+                         "Neo.ClientError.Transaction.InvalidBookmarkMixture"):
+            return True
+        if (self.code.startswith("Neo.ClientError.Security.")
+                and self.code != "Neo.ClientError.Security."
+                                 "AuthorizationExpired"):
+            return True
+        return False
+
     def __str__(self):
         return "{{code: {code}}} {{message: {message}}}".format(code=self.code, message=self.message)
 
@@ -136,6 +152,8 @@ class Neo4jError(Exception):
 class ClientError(Neo4jError):
     """ The Client sent a bad request - changing the request might yield a successful outcome.
     """
+    def __str__(self):
+        return super(Neo4jError, self).__str__()
 
 
 class DatabaseError(Neo4jError):
@@ -199,6 +217,12 @@ class AuthError(ClientError):
     """
 
 
+class TokenExpired(AuthError):
+    """ Raised when the authentication token has expired.
+
+    A new driver instance with a fresh authentication token needs to be created.
+    """
+
 client_errors = {
 
     # ConstraintError
@@ -228,8 +252,11 @@ client_errors = {
     "Neo.ClientError.Security.AuthorizationFailed": AuthError,
     "Neo.ClientError.Security.Unauthorized": AuthError,
 
+    # TokenExpired
+    "Neo.ClientError.Security.TokenExpired": TokenExpired,
+
     # NotALeader
-    "Neo.ClientError.Cluster.NotALeader": NotALeader
+    "Neo.ClientError.Cluster.NotALeader": NotALeader,
 }
 
 transient_errors = {
