@@ -98,13 +98,6 @@ class Bolt4x0(Bolt):
     def der_encoded_server_certificate(self):
         return self.socket.getpeercert(binary_form=True)
 
-    @property
-    def local_port(self):
-        try:
-            return self.socket.getsockname()[1]
-        except OSError:
-            return 0
-
     def get_base_headers(self):
         return {
             "user_agent": self.user_agent,
@@ -265,6 +258,10 @@ class Bolt4x0(Bolt):
         self.send_all()
         self.fetch_all()
 
+    def goodbye(self):
+        log.debug("[#%04X]  C: GOODBYE", self.local_port)
+        self._append(b"\x02", ())
+
     def fetch_message(self):
         """ Receive at most one message from the server, if available.
 
@@ -325,25 +322,6 @@ class Bolt4x0(Bolt):
 
         return len(details), 1
 
-    def close(self):
-        """ Close the connection.
-        """
-        if not self._closed:
-            if not self._defunct:
-                log.debug("[#%04X]  C: GOODBYE", self.local_port)
-                self._append(b"\x02", ())
-                try:
-                    self._send_all()
-                except (OSError, BoltError, DriverError):
-                    pass
-            log.debug("[#%04X]  C: <CLOSE>", self.local_port)
-            try:
-                self.socket.close()
-            except OSError:
-                pass
-            finally:
-                self._closed = True
-
     def closed(self):
         return self._closed
 
@@ -390,6 +368,20 @@ class Bolt4x3(Bolt4x2):
 
     PROTOCOL_VERSION = Version(4, 3)
 
+    def get_base_headers(self):
+        """ Bolt 4.1 passes the routing context, originally taken from
+        the URI, into the connection initialisation message. This
+        enables server-side routing to propagate the same behaviour
+        through its driver.
+        """
+        headers = {
+            "user_agent": self.user_agent,
+            "patch_bolt": ["utc"]
+        }
+        if self.routing_context is not None:
+            headers["routing"] = self.routing_context
+        return headers
+
     def route(self, database=None, imp_user=None, bookmarks=None):
         if imp_user is not None:
             raise ConfigurationError(
@@ -416,6 +408,7 @@ class Bolt4x3(Bolt4x2):
 
     def hello(self):
         def on_success(metadata):
+            # configuration hints
             self.configuration_hints.update(metadata.pop("hints", {}))
             self.server_info.update(metadata)
             if "connection.recv_timeout_seconds" in self.configuration_hints:
@@ -429,6 +422,8 @@ class Bolt4x3(Bolt4x2):
                              "connection.recv_timeout_seconds (%r). Make sure "
                              "the server and network is set up correctly.",
                              self.local_port, recv_timeout)
+            # bolt patch handshake
+            self.bolt_patches.update(set(metadata.pop("patch_bolt", ())))
 
         headers = self.get_base_headers()
         headers.update(self.auth_dict)
