@@ -23,51 +23,11 @@ from collections import deque
 from warnings import warn
 
 from neo4j.data import DataDehydrator
-from neo4j.exceptions import (
-    ServiceUnavailable,
-    SessionExpired,
-)
+from neo4j.io import ConnectionErrorHandler
+from neo4j.meta import experimental
 from neo4j.work.summary import ResultSummary
 
 
-class _ConnectionErrorHandler:
-    """
-    Wrapper class for handling connection errors.
-
-    The class will wrap each method to invoke a callback if the method raises
-    SessionExpired or ServiceUnavailable.
-    The error will be re-raised after the callback.
-    """
-
-    def __init__(self, connection, on_network_error):
-        """
-        :param connection the connection object to warp
-        :type connection Bolt
-        :param on_network_error the function to be called when a method of
-            connection raises of of the caught errors. The callback takes the
-            error as argument.
-        :type on_network_error callable
-        """
-        self._connection = connection
-        self._on_network_error = on_network_error
-
-    def __getattr__(self, item):
-        connection_attr = getattr(self._connection, item)
-        if not callable(connection_attr):
-            return connection_attr
-
-        def outer(func):
-            def inner(*args, **kwargs):
-                try:
-                    func(*args, **kwargs)
-                finally:
-                    if self._connection.defunct():
-                        self._on_network_error()
-            return inner
-
-        return outer(connection_attr)
-
-
 class Result:
     """A handler for the result of Cypher query execution. Instances
     of this class are typically constructed and returned by
@@ -75,15 +35,15 @@ class Result:
     """
 
     def __init__(self, connection, hydrant, fetch_size, on_closed,
-                 on_network_error):
-        self._connection = _ConnectionErrorHandler(connection, on_network_error)
+                 on_error):
+        self._connection = ConnectionErrorHandler(connection, on_error)
         self._hydrant = hydrant
         self._on_closed = on_closed
         self._metadata = None
         self._record_buffer = deque()
         self._summary = None
         self._bookmark = None
-        self._qid = -1
+        self._raw_qid = -1
         self._fetch_size = fetch_size
 
         # states
@@ -96,18 +56,29 @@ class Result:
         # the result has been fully iterated or consumed
         self._closed = False
 
+    @property
+    def _qid(self):
+        if self._raw_qid == self._connection.most_recent_qid:
+            return -1
+        else:
+            return self._raw_qid
+
     def _tx_ready_run(self, query, parameters, **kwparameters):
         # BEGIN+RUN does not carry any extra on the RUN message.
         # BEGIN {extra}
         # RUN "query" {parameters} {extra}
-        self._run(query, parameters, None, None, None, **kwparameters)
+        self._run(query, parameters, None, None, None, None, **kwparameters)
 
-    def _run(self, query, parameters, db, access_mode, bookmarks, **kwparameters):
+    def _run(self, query, parameters, db, imp_user, access_mode, bookmarks,
+             **kwparameters):
         query_text = str(query)  # Query or string object
         query_metadata = getattr(query, "metadata", None)
         query_timeout = getattr(query, "timeout", None)
 
-        parameters = DataDehydrator.fix_parameters(dict(parameters or {}, **kwparameters))
+        parameters = DataDehydrator.fix_parameters(
+            dict(parameters or {}, **kwparameters),
+            patch_utc="utc" in self._connection.bolt_patches
+        )
 
         self._metadata = {
             "query": query_text,
@@ -117,7 +88,10 @@ class Result:
 
         def on_attached(metadata):
             self._metadata.update(metadata)
-            self._qid = metadata.get("qid", -1)  # For auto-commit there is no qid and Bolt 3 do not support qid
+            # For auto-commit there is no qid and Bolt 3 does not support qid
+            self._raw_qid = metadata.get("qid", -1)
+            if self._raw_qid != -1:
+                self._connection.most_recent_qid = self._raw_qid
             self._keys = metadata.get("fields")
             self._attached = True
 
@@ -134,6 +108,7 @@ class Result:
             metadata=query_metadata,
             timeout=query_timeout,
             db=db,
+            imp_user=imp_user,
             on_success=on_attached,
             on_failure=on_failed_attach,
         )
@@ -223,20 +198,37 @@ class Result:
         self._closed = True
 
     def _attach(self):
-        """Sets the Result object in an attached state by fetching messages from the connection to the buffer.
+        """Sets the Result object in an attached state by fetching messages from
+        the connection to the buffer.
         """
         if self._closed is False:
             while self._attached is False:
                 self._connection.fetch_message()
 
-    def _buffer_all(self):
-        """Sets the Result object in an detached state by fetching all records from the connection to the buffer.
+    def _buffer(self, n=None):
+        """Try to fill `self_record_buffer` with n records.
+
+        Might end up with more records in the buffer if the fetch size makes it
+        overshoot.
+        Might ent up with fewer records in the buffer if there are not enough
+        records available.
         """
         record_buffer = deque()
         for record in self:
             record_buffer.append(record)
+            if n is not None and len(record_buffer) >= n:
+                break
         self._closed = False
-        self._record_buffer = record_buffer
+        if n is None:
+            self._record_buffer = record_buffer
+        else:
+            self._record_buffer.extend(record_buffer)
+
+    def _buffer_all(self):
+        """Sets the Result object in an detached state by fetching all records
+        from the connection to the buffer.
+        """
+        self._buffer()
 
     def _obtain_summary(self):
         """Obtain the summary of this result, buffering any remaining records.
@@ -245,9 +237,14 @@ class Result:
         """
         if self._summary is None:
             if self._metadata:
-                self._summary = ResultSummary(**self._metadata)
+                self._summary = ResultSummary(
+                    self._connection.unresolved_address, **self._metadata
+                )
             elif self._connection:
-                self._summary = ResultSummary(server=self._connection.server_info)
+                self._summary = ResultSummary(
+                    self._connection.unresolved_address,
+                    server=self._connection.server_info
+                )
 
         return self._summary
 
@@ -268,11 +265,11 @@ class Result:
                 result = tx.run("CREATE (n:ExampleNode { name: $name }) RETURN n", name=name)
                 record = result.single()
                 value = record.value()
-                info = result.consume()
-                return value, info
+                summary = result.consume()
+                return value, summary
 
             with driver.session() as session:
-                node_id, info = session.write_transaction(create_node_tx, "example")
+                node_id, summary = session.write_transaction(create_node_tx, "example")
 
         Example::
 
@@ -283,12 +280,12 @@ class Result:
                     if x > 1:
                         break
                     values.append(record.values())
-                info = result.consume()  # discard the remaining records if there are any
-                # use the info for logging etc.
-                return values, info
+                summary = result.consume()  # discard the remaining records if there are any
+                # use the summary for logging etc.
+                return values, summary
 
             with driver.session() as session:
-                values, info = session.read_transaction(get_two_tx)
+                values, summary = session.read_transaction(get_two_tx)
 
         :returns: The :class:`neo4j.ResultSummary` for this result
         """
@@ -309,6 +306,13 @@ class Result:
         :returns: the next :class:`neo4j.Record` or :const:`None` if none remain
         :warns: if more than one record is available
         """
+        # TODO in 5.0 replace with this code that raises an error if there's not
+        # exactly one record in the left result stream.
+        # self._buffer(2).
+        # if len(self._record_buffer) != 1:
+        #     raise SomeError("Expected exactly 1 record, found %i"
+        #                      % len(self._record_buffer))
+        # return self._record_buffer.popleft()
         records = list(self)  # TODO: exhausts the result with self.consume if there are more records.
         size = len(records)
         if size == 0:
@@ -321,18 +325,11 @@ class Result:
         """Obtain the next record from this result without consuming it.
         This leaves the record in the buffer for further processing.
 
-        :returns: the next :class:`.Record` or :const:`None` if none remain
+        :returns: the next :class:`neo4j.Record` or :const:`None` if none remain
         """
+        self._buffer(1)
         if self._record_buffer:
             return self._record_buffer[0]
-        if not self._attached:
-            return None
-        while self._attached:
-            self._connection.fetch_message()
-            if self._record_buffer:
-                return self._record_buffer[0]
-
-        return None
 
     def graph(self):
         """Return a :class:`neo4j.graph.Graph` instance containing all the graph objects
@@ -341,6 +338,8 @@ class Result:
 
         :returns: a result graph
         :rtype: :class:`neo4j.graph.Graph`
+
+        **This is experimental.** (See :ref:`filter-warnings-ref`)
         """
         self._buffer_all()
         return self._hydrant.graph
@@ -378,3 +377,28 @@ class Result:
         :rtype: list
         """
         return [record.data(*keys) for record in self]
+
+    @experimental("pandas support is experimental and might be changed or "
+                  "removed in future versions")
+    def to_df(self):
+        """Convert (the rest of) the result to a pandas DataFrame.
+
+        This method is only available if the `pandas` library is installed.
+
+        ``tx.run("UNWIND range(1, 10) AS n RETURN n, n+1 as m").to_df()``, for
+        instance will return a DataFrame with two columns: ``n`` and ``m`` and
+        10 rows.
+
+        :rtype: :py:class:`pandas.DataFrame`
+        :raises ImportError: if `pandas` library is not available.
+
+        .. versionadded:: 5.0
+            This method was backported from 5.0 for preview purposes.
+
+        **This is experimental.**
+        ``pandas`` support might be changed or removed in future versions
+        without warning. (See :ref:`filter-warnings-ref`)
+        """
+        import pandas as pd
+
+        return pd.DataFrame(self.values(), columns=self._keys)
