@@ -206,7 +206,7 @@ class Session(Workspace):
         protocol_version = cx.PROTOCOL_VERSION
         server_info = cx.server_info
 
-        hydrant = DataHydrator()
+        hydrant = DataHydrator(patch_utc="utc" in cx.bolt_patches)
 
         self._autoResult = Result(
             cx, hydrant, self._config.fetch_size, self._result_closed,
@@ -355,6 +355,11 @@ class Session(Workspace):
 
     def read_transaction(self, transaction_function, *args, **kwargs):
         """Execute a unit of work in a managed read transaction.
+
+        .. note::
+            This does not necessarily imply access control, see the session
+            configuration option :ref:`default-access-mode-ref`.
+
         This transaction will automatically be committed unless an exception is thrown during query execution or by the user code.
         Note, that this function perform retries and that the supplied `transaction_function` might get invoked more than once.
 
@@ -381,8 +386,9 @@ class Session(Workspace):
                     if x > 1:
                         break
                     values.append(record.values())
-                info = result.consume()  # discard the remaining records if there are any
-                # use the info for logging etc.
+                # discard the remaining records if there are any
+                summary = result.consume()
+                # use the summary for logging etc.
                 return values
 
             with driver.session() as session:
@@ -397,6 +403,11 @@ class Session(Workspace):
 
     def write_transaction(self, transaction_function, *args, **kwargs):
         """Execute a unit of work in a managed write transaction.
+
+        .. note::
+            This does not necessarily imply access control, see the session
+            configuration option :ref:`default-access-mode-ref`.
+
         This transaction will automatically be committed unless an exception is thrown during query execution or by the user code.
         Note, that this function perform retries and that the supplied `transaction_function` might get invoked more than once.
 
@@ -429,7 +440,7 @@ class Query:
     :param metadata: metadata attached to the query.
     :type metadata: dict
     :param timeout: seconds.
-    :type timeout: int
+    :type timeout: float or :const:`None`
     """
     def __init__(self, text, metadata=None, timeout=None):
         self.text = text
@@ -446,6 +457,8 @@ def unit_of_work(metadata=None, timeout=None):
 
     For example, a timeout may be applied::
 
+        from neo4j import unit_of_work
+
         @unit_of_work(timeout=100)
         def count_people_tx(tx):
             result = tx.run("MATCH (a:Person) RETURN count(a) AS persons")
@@ -465,7 +478,7 @@ def unit_of_work(metadata=None, timeout=None):
         This functionality allows to limit query/transaction execution time.
         Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
         Value should not represent a duration of zero or negative duration.
-    :type timeout: int
+    :type timeout: float or :const:`None`
     """
 
     def wrapper(f):
